from __future__ import annotations
from dataclasses import dataclass
import random
import pandas as pd
from abc import ABC, abstractmethod


@dataclass
class Static_data_configuration_level:
    name: str
    feature: DataType


class DATA_TYPES:
    RANGE = "range"
    CHOICE = "choice"


class DataType(ABC):
    kind: str

    @abstractmethod
    def get_value(self):
        """Generates a value for this data type

        Returns
        -------
        [type]
            Value from the given type
        """

    @abstractmethod
    def get_values(self, n) -> list:
        """Generates `n` values for this data type

        Parameters
        ----------
        n : int
            Number of values to generate

        Returns
        -------
        list
            `n` random values"""


class RangeDataType(DataType):
    kind = "range"

    def __init__(
        self, max_range, min_range=0, as_integer: bool = False, **kwargs
    ) -> None:
        self.min_range = min_range
        self.max_range = max_range
        self.as_integer = as_integer
        self._delta = max_range - min_range

    def _get_rand_int(self):
        return random.randrange(
            self.min_range,
            self.max_range,
        )

    def _get_rand_float(self):
        return self.min_range + (random.random() * self._delta)

    def get_value(self):
        if self.as_integer:
            return self._get_rand_int()
        return self._get_rand_float()

    def get_values(self, n: int) -> list:
        return [self.get_value() for _ in range(n)]

    def to_dict(self):
        return {
            "kind": self.kind,
            "min_range": self.min_range,
            "max_range": self.max_range,
            "as_integer": self.as_integer,
        }


class ChoiceDataType(DataType):
    kind = "choice"

    def __init__(self, choices: list, **kwargs) -> None:
        self.choices = choices

    def get_value(self):
        return random.choice(self.choices)

    def get_values(self, n: int) -> list:
        return random.choices(self.choices, k=n)

    def to_dict(self) -> dict:
        return {"kind": self.kind, "choices": self.choices}


def get_data_type(level_configuration: dict) -> DataType:
    """DataType Factory according to provided `kind`.
    possible DataTypes:
    - Range
    - Choice

    Parameters
    ----------
    level_configuration : dict
        Static feature configuration

    Returns
    -------
    DataType
        DataType representing the configured feature

    Raises
    ------
    ValueError
        Provided kind is not in the available DATA_TYPES
    """
    kind = level_configuration.get("kind")
    if kind == DATA_TYPES.RANGE:
        return RangeDataType(**level_configuration)
    elif kind == DATA_TYPES.CHOICE:
        return ChoiceDataType(**level_configuration)
    raise ValueError(
        f"Configuration {kind} is not available, try {list(DATA_TYPES.__members__.keys())}"
    )


class Static_data_generator:
    def __init__(self, deployment: pd.DataFrame, configuration: dict) -> None:
        """Creates a static data generator according to givein configuration

        Parameters
        ----------
        deployment : pd.DataFrame
            Deployment dataframe, generated by the `deployment_generator`
        configuration : dict
            Configuration dictionary of static data or containing such at `static` entry
            Each entry can hold:
            - A list of optional values ex: [model_a, model_b, model_c], [1, 3, 5, 7, 11, 13]
            - A range string ex: range(10)
        """
        self.deployment = deployment.copy(deep=True)
        self.configuration: list = self._get_or_create_static_data_configuration(
            configuration
        )

    def _get_or_create_static_data_configuration(
        self, static_data_configuration: dict
    ) -> list:
        """Converts a static data configuration dictionary to a list of `Static_data_configuration_level`

        Parameters
        ----------
        static_data_configuration : dict
            Static data configuration

        Returns
        -------
        list[Static_data_configuration_level]
            a list of parsed `Static_data_configuration_level`
        """
        configuration = []
        if "static" in static_data_configuration:
            static_data_configuration = static_data_configuration["static"]
        for feature_name, feature_configuration in static_data_configuration.items():
            try:
                feature = get_data_type(feature_configuration)
                configuration.append(
                    Static_data_configuration_level(feature_name, feature)
                )
            except ValueError as e:
                continue
        return configuration

    def generate_static_data(self) -> pd.DataFrame:
        """Generates a dataframe with static data for the provided deployment
        according to the given configuration

        Returns
        -------
        pd.DataFrame
            DataFrame that contains the generated static data
        """
        deployment = self.deployment.copy()
        for static_feature in self.configuration:
            deployment[static_feature.name] = static_feature.feature.get_values(
                self.deployment.shape[0]
            )
        return deployment



import pandas as pd
from metrics_gen.metric.metric_group import Metric_Group
import hashlib
import datetime
from pytimeparse import parse
import json
import itertools


class Generator_df:
    def __init__(
        self,
        configuration: dict,
        initial_timestamp=datetime.datetime.now(),
        user_hierarchy=pd.DataFrame,
    ):
        """Metrics factory"""
        metrics = configuration["metrics"]
        timestamps = configuration["timestamps"]
        errors = configuration["errors"]

        self.dimension_hashes = []
        self.initial_values = {}

        for metric in metrics:
            if metric not in user_hierarchy.columns:
                user_hierarchy[metric] = 0

        self.dimensions = self.define_uh(
            df_labels=user_hierarchy[
                list(list(user_hierarchy.columns) - metrics.keys())
            ],
            df_metrics=user_hierarchy[metrics.keys()],
        )

        self.metric_groups = {
            dimension_hash: Metric_Group(
                metrics=metrics,
                initial_values=self.initial_values[dimension_hash],
                error_length_ticks=errors.setdefault("length_in_ticks", 0),
                error_rate_ticks=errors.setdefault("rate_in_ticks", 0),
            )
            for dimension_hash in self.dimension_hashes
        }

        self.global_timestamp = self.get_timestamp(initial_timestamp)
        self.timestamp_interval = self.get_interval(
            timestamps.setdefault("interval", "1s")
        )
        self.indexes = ["timestamp", [label for label in self.metric_groups.keys()]]

    def get_timestamp(self, user_timestamp):
        if type(user_timestamp) == datetime.datetime:
            return user_timestamp
        else:
            return datetime.datetime.fromtimestamp(user_timestamp)

    def get_interval(self, user_interval):
        if type(user_interval) == int:
            return datetime.timedelta(seconds=user_interval)
        elif type(user_interval) == str:
            seconds = parse(user_interval)
            return datetime.timedelta(seconds=seconds)
        else:
            return user_interval

    def define_uh(self, df_labels: pd.DataFrame, df_metrics: pd.DataFrame):
        """[summary]

        Parameters
        ----------
        df_labels : pd.DataFrame
            Deployment columns part in ther user DF
        df_metrics : pd.DataFrame
            Metric columns part in the user DF

        Returns
        -------
        [type]
            [description]
        """
        # Extract labels
        labels = []
        for idx, cols in df_labels.iterrows():
            current_label = cols.to_dict().items()
            labels.append(
                (
                    list(current_label),
                    json.loads(df_metrics.iloc[idx].to_json(orient="index")),
                )
            )

        # hash labels
        for label in labels:
            row_labels_hash = hashlib.sha256(str(label[0]).encode()).hexdigest()
            self.dimension_hashes.append(row_labels_hash)
            self.initial_values[row_labels_hash] = label[1]
        return list(map(lambda l: l[0], labels))

    def get_dataframe_hash(self, df: pd.DataFrame):
        return hashlib.sha256(df.to_json().encode()).hexdigest()

    def get_new_timestamp(self):
        self.global_timestamp += self.timestamp_interval
        return self.global_timestamp

    def get_metric_groups_values_as_dict(self):
        res = []
        sample_timestamp = self.get_new_timestamp()
        for dimension in self.dimensions:
            dimension_name = hashlib.sha256(str(dimension).encode()).hexdigest()
            metric_group = self.metric_groups[dimension_name]
            metrics = next(metric_group.generate())
            metrics["timestamp"] = sample_timestamp
            res.append([*dimension, *metrics.items()])
        return res

    def build_dict_from_tuples_array(self, tuples_array: list):
        for row in tuples_array:
            for key, value in row:
                # is it a label?
                if (key, value) in self.dimensions:
                    print("*")
        return tuples_array

    def transform_metric_groups_dict_to_df(self, metric_groups_dict: dict):
        df = pd.DataFrame()

        # Add columns
        for mg in metric_groups_dict:
            mg_dict = dict(mg)
            row = pd.DataFrame(data=mg_dict, index=[0])
            df = df.append(row)
        # Add total is_error
        df["is_error"] = df.filter(regex=".*is_error.*").apply(
            lambda row: row.min(), axis=1
        )

        # Add timestamp
        df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
        new_indexes = ["timestamp", *list(map(lambda v: v[0], self.dimensions[0]))]
        df = df.set_index(new_indexes)
        return df

    def generate(self, as_df: bool = True):
        while True:
            res = self.get_metric_groups_values_as_dict()
            if as_df:
                res = self.transform_metric_groups_dict_to_df(res)
            else:
                res = self.build_dict_from_tuples_array(res)
            yield res

    def _range_iterator_crator(
        self,
        start_time=datetime.datetime.now(),
        end_time=datetime.datetime.now() + datetime.timedelta(seconds=1),
        as_df: bool = True,
    ):
        self.global_timestamp = start_time
        gen = self.generate(as_df)
        if end_time >= self.global_timestamp:
            while end_time >= self.global_timestamp:
                yield next(gen)
        else:
            while True:
                yield pd.DataFrame()

    def generate_range(
        self,
        start_time=datetime.datetime.now(),
        end_time=datetime.datetime.now() + datetime.timedelta(seconds=1),
        as_df: bool = True,
        as_iterator=True,
    ):
        self.global_timestamp = start_time
        if as_df and not as_iterator:
            metrics = self._range_iterator_crator(start_time, end_time, False)
        else:
            metrics = self._range_iterator_crator(start_time, end_time, as_df)

        if as_iterator:
            return metrics
        else:
            exhausted_metrics = list(map(lambda x: dict(x), itertools.chain(*metrics)))
            if as_df:
                return pd.DataFrame.from_dict(exhausted_metrics)
            else:
                return exhausted_metrics



from metrics_gen.static_data_generator import Static_data_generator
from metrics_gen.deployment_generator import deployment_generator
import pandas as pd
import yaml


def get_deployment(configuration: dict) -> pd.DataFrame:
    dep_gen = deployment_generator()
    deployment = dep_gen.generate_deployment(configuration=configuration)
    return deployment


class TestStaticData:
    metrics_configuration: dict = yaml.safe_load(
        open(
            "./tests/test_configuration.yaml",
            "r",
        )
    )
    static_configuration: dict = metrics_configuration.get("static", {})
    deployment: pd.DataFrame = get_deployment(metrics_configuration)

    def test_create_generator_with_full_configuration(self):
        static_data_generator = Static_data_generator(
            self.deployment, self.metrics_configuration
        )
        assert (
            static_data_generator,
            "Static data generator failed to be created from a full file configuration",
        )

    def test_create_generator_with_static_configuration(self):
        static_data_generator = Static_data_generator(
            self.deployment, self.static_configuration
        )
        assert (
            static_data_generator,
            "Static data generator failed to be created from static configuration",
        )

    def test_static_data_created(self):
        static_data_generator = Static_data_generator(
            self.deployment, self.static_configuration
        )

        generated_df = static_data_generator.generate_static_data()
        assert (
            not generated_df.equals(self.deployment),
            "No static data was generated, returned original deployment DF",
        )



import itertools
import pandas as pd
import faker
from .deployment import providers
from dataclasses import dataclass


@dataclass
class Deployment_level:
    name: str
    num_items: int
    level_type: faker.providers.BaseProvider


class deployment_generator:
    def __init__(self):
        """Creates a deployment factory"""
        self.faker = faker.Faker("en_US")

        self.faker.add_provider(providers.LocationProvider)

        self.configuration: list = list()
        self.temp_location_configuration: dict = dict()

    def generate_deployment(self, configuration: dict = dict()) -> pd.DataFrame:
        """Generates a deployment

        Parameters
        ----------
        configuration : list, optional
            A deployment configuration, if not provided assumes `self.add_levels()` was used
            beforehand to create a levels configuration

        Returns
        -------
        pd.DataFrame
            Generated deployment
        """

        # Gather deployment configurations
        deployment_configuration = self._get_or_create_configuration(configuration)

        # Populate levels with data
        tmp_generation = self._add_column_to_sample([], deployment_configuration.copy())
        columns = self._extract_columns_from_configuration(deployment_configuration)

        # Create deployment DF
        df = pd.DataFrame(data=tmp_generation, columns=columns)

        if not self.temp_location_configuration:
            return df
        else:
            return self._add_location_to_df(df)

    def get_faker(self):
        """Returns the currently configured faker instance"""
        return self.faker

    def add_location(self, location_level: str, location_bounds: dict):
        """Add a location parameter to the deployment (as the last level)

        Parameters
        ----------
        location_level : str
            The deployment level to add the location to
        location_bounds : dict
            The box bounds of the location, as a dict containing the `nw` (north-west)
            and the `se` (south-east) boundary points.
        """
        self.temp_location_configuration = {
            "level": location_level,
            "nw": location_bounds["nw"],
            "se": location_bounds["se"],
        }

    def _get_or_create_configuration(self, configuration: dict = dict()) -> dict:
        # Do we need to create levels from ad-hoc configuration?
        if configuration:
            # Need to parse the current configuration and add the levels accordingly
            deployment_configuration = self._get_deployment_config(configuration)

            # Create levels according to configuration
            for level, level_configs in deployment_configuration.items():
                devices_to_generate = level_configs.get("num_items", 1)
                level_type = level_configs.get("faker", "msisdn")
                self.add_level(
                    level,
                    number=devices_to_generate,
                    level_type=getattr(self.faker, level_type),
                )
        deployment_configuration = self.configuration
        return deployment_configuration

    @staticmethod
    def _get_deployment_config(config: dict) -> dict:
        """Returns the  static data part from the config.
        if there is no `static` entry in the provided config
        it will assume the entire provided config is a static data config
        config.

        Parameters
        ----------
        config : dict
            A static data configuration, or a configuration containing `static` entry
            for a static data featuers configuration

        Returns
        -------
        dict
            static data features configuration
        """
        return config.get("static", config)

    @staticmethod
    def _get_deployment_config(config: dict) -> dict:
        """Returns the  deployment part from the config.
        if there is no `deployment` entry in the provided config
        it will assume the entire provided config is a deployment
        config.

        Parameters
        ----------
        config : dict
            A Deployment configuration, or a configuration containing `deployment` entry
            for a deployment configuration

        Returns
        -------
        dict
            deployment configuration
        """
        return config.get("deployment", config)

    def _add_location_to_df(self, df: pd.DataFrame):
        level = self._get_location_level()
        bounding_box = self._get_location_bounding_box()
        unique_level_values = df[level].unique()
        locations = {
            current_value: str(self.faker.location(bounding_box))
            for current_value in unique_level_values
        }
        df["location"] = df[level].apply(lambda val: locations[val])
        return df

    def _get_location_level(self):
        return self.temp_location_configuration["level"]

    def _get_location_bounding_box(self):
        return {
            "nw": self.temp_location_configuration["nw"],
            "se": self.temp_location_configuration["se"],
        }

    def add_level(self, name: str, number: int, level_type):
        """Adds a deployment level to the deployment generator.
        Deployment levels allows us to have hierarchy such as:
        Data Center -> Device,
        For each level, `number` of instances will be created.

        So for a 2 data_centers, 2 devices deployment, 4 devices will be
        created, 2 devices in each of the 2 data centers.

        Parameters
        ----------
        name : str
            Level name
        number : int
            Number of instances to create (This `number` will be created for
            each instance of the previous levels)
        level_type : str
            Name of the faker generator to be used for generating the
            level's instances
        """
        self.configuration.append(Deployment_level(name, number, level_type))

    def _add_config_name(self, name: str):
        return 0

    def _add_config_number(self, num: int):
        return 0

    def _is_data_generation_needed(self, level_tuple: tuple):
        if type(level_tuple[1]) == int:
            return True
        return False

    def _extract_columns_from_configuration(self, configuration: list):
        return list(map(lambda levels: levels.name, configuration))

    def _add_column_to_sample(self, current: list, left: list):
        """Generates the data for each column as defined in the list

        Parameters
        ----------
        current : list
            [description]
        left : list
            [description]

        Returns
        -------
        [type]
            [description]
        """
        current_level: Deployment_level = left.pop(0)
        current_generator = current_level.level_type
        num_possibilities_to_generate = current_level.num_items
        to_append = [
            current_generator().replace(",", "_").replace(" ", "_")
            for _ in range(num_possibilities_to_generate)
        ]
        if not left:
            generated = [[*current, elem] for elem in to_append]
            return generated

        generated = [
            self._add_column_to_sample([*current, elem], left.copy())
            for elem in to_append
        ]
        return list(itertools.chain.from_iterable(generated))



from random import Random
from metrics_gen.metric.metrics import normal
from metrics_gen.metric.metrics import poisson
import numpy as np


class Metric:
    """
    - Define metric parameters
    - Define distribution parameters
    - Define alerts
    """

    def __init__(self, configuration: dict = dict(), initial_value: float = 0):
        # General
        self.r = Random()
        # self.name = configuration.setdefault('name', f'metric_r{r.randint(0, 1000)}')
        self.available_distributions = {
            "normal": normal.Normal,
            "poisson": poisson.Poisson,
        }

        # Error configurations
        self.is_threshold_below = configuration.setdefault("is_threshold_below", True)
        self.is_in_peak_error = False
        self.is_error = False
        self.steps = 0

        self.error_length = 0
        self.peaks = 0
        self.error_peak_length = 0
        self.error_metric = self.Peak_error()
        self.peak_chance = 0

        self.value_history = initial_value
        self.validations = configuration.setdefault(
            "validation", {"distribution": {}, "metric": {}}
        )
        self.validation_distribution = self.validations["distribution"]
        self.validation_metric = self.validations["metric"]
        self.should_validate_distribution_values = (
            self.validation_distribution.setdefault("validate", False)
        )
        self.should_validate_metric_values = self.validation_metric.setdefault(
            "validate", False
        )

        # Distribution
        self.distribution = self.available_distributions[
            configuration.setdefault("distribution", "normal")
        ]
        self.params = configuration.setdefault("distribution_params", {})
        self.past_based_value = configuration.setdefault("past_based_value", False)

        # Validation
        self.distribution_min = self.validation_distribution.setdefault("min", np.NINF)
        self.distribution_max = self.validation_distribution.setdefault("max", np.Inf)

        self.metric_min = self.validation_metric.setdefault("min", np.NINF)
        self.metric_max = self.validation_metric.setdefault("max", np.Inf)

        if self.should_validate_distribution_values:
            self.validate_min_max("distribution")
        if self.should_validate_metric_values:
            self.validate_min_max("metric")

    def Peak_error(self, target_peaks: int = 4, error_peak_ratio: float = 0.5):
        def return_peak():
            return self.metric_max if self.is_threshold_below else self.metric_min

        while True:
            # Are we in Pre-Error?
            if self.steps <= self.error_length - self.error_peak_length:
                # Will it peak?
                is_peak = True if self.r.uniform(0, 1) <= self.peak_chance else False
                yield return_peak() if is_peak else self.distribution(**self.params)[0]

            # Are we in Peak-Error?
            else:
                self.is_in_peak_error = True
                yield return_peak()

    def start_error(
        self, error_length: int, target_peaks: int = 4, error_peak_ratio: float = 0.5
    ) -> None:
        r = Random()

        # Pick one error scenario
        self.error_length = error_length
        self.is_error = True
        self.error_metric = self.Peak_error()
        self.peaks = int(r.gauss(mu=target_peaks, sigma=0.5 * target_peaks))
        self.error_peak_length = int(
            r.gauss(
                mu=self.error_length * error_peak_ratio, sigma=self.error_length * 0.1
            )
        )
        self.peak_chance = self.peaks / (self.error_length - self.error_peak_length)

    def stop_error(self) -> None:
        # Return generator to Normal
        self.current_metric = self.distribution
        self.is_error = False
        self.is_in_peak_error = False
        self.steps = 0
        self.error_length = 0

    def get_value(self):
        """
            Produces the metric from a distribution as defined by the user
        :return: One metric sample
        """

        if self.is_error:
            self.steps += 1
            return next(self.error_metric)
        else:
            new_value = self.distribution(**self.params)[0]
            new_value = (
                self.validate_value("distribution", new_value)
                if self.should_validate_distribution_values
                else new_value
            )
            if self.past_based_value:
                self.value_history += new_value
                return self.value_history
            else:
                return new_value

    def validate_value(self, validation_type: str, value):
        """
        Validates the values are within valid range as defined by min / max
        """
        min = (
            self.distribution_min
            if (validation_type == "distribution")
            else self.metric_min
        )
        max = (
            self.distribution_max
            if (validation_type == "distribution")
            else self.metric_max
        )
        value = value if value > min else min
        value = value if value < max else max
        return value

    def validate_min_max(self, validation_type: str):
        min = (
            self.distribution_min
            if validation_type == "distribution"
            else self.metric_min
        )
        max = (
            self.distribution_max
            if validation_type == "distribution"
            else self.metric_max
        )
        if min > max:
            raise ValueError(
                f"In metric {self.name}, max < min, please update configuration"
            )

    def generator(self):
        while True:
            value = self.get_value()
            res = (
                self.validate_value("metric", value)
                if self.should_validate_metric_values
                else value
            )
            yield {"value": res, "is_error": self.is_in_peak_error}



import numpy as np


def Poisson(lam=2, noise=0):
    """
    The function used to create the actual device performance data

    :param lambda: Mean of the normal distribution to draw the metric from
    :param noise: Added noise, (Drawn from distribution of Normal(0, noise)
    :return: Metric
    """
    # while True:
    added_noise = 0
    if noise is not 0:
        added_noise = np.random.normal(loc=0, scale=noise, size=1)
    tick = np.random.poisson(lam=lam, size=1) + added_noise
    return tick



import numpy as np


def Normal(mu=0, sigma=0.5, noise=0):
    """
    The function used to create the actual device performance data

    :param mu: Mean of the normal distribution to draw the metric from
    :param sigma: Deviation from the mean of the distribution
    :param noise: Sigma for the noise, (Drawn from distribution of Normal(0, noise)
    :return: Metric
    """
    # while True:
    added_noise = 0
    if noise is not 0:
        added_noise = np.random.normal(loc=0, scale=noise, size=1)
    tick = np.random.normal(loc=mu, scale=sigma, size=1) + added_noise
    return tick



from random import Random
from metrics_gen.metric.metric import Metric


class Metric_Group:
    def __init__(
        self,
        metrics: dict,
        initial_values: dict = {},
        error_rate_ticks: int = 0,
        error_length_ticks: int = 0,
    ):
        """
            Component Manager:
            Receives configuration dictionary and -
                - Creates metrics
                - Runs scenarios
        :param metrics: Configuration dictionary
        """

        # Error handling
        self.error_rate_in_percentage = 1 / error_rate_ticks
        self.error_length_in_ticks = error_length_ticks
        self.is_error = False
        self.steps = 0
        self.current_error_length = 0
        self.scenario = {}

        # Metrics definition
        self.metrics = {
            metric_name: Metric(
                metric_config, initial_value=initial_values.setdefault(metric_name, 0)
            )
            for metric_name, metric_config in metrics.items()
        }

        self.r = Random()

        self.total_steps = 0

    def notify_metric_of_error(self) -> None:
        for metric_name, component in self.metrics.items():
            if self.steps == self.scenario[metric_name]:
                component.start_error(self.error_length_in_ticks - self.steps)

    def notify_metrics_of_normalization(self):
        for component in self.metrics.values():
            component.stop_error()

    def generate(self):
        # Initialize state

        # Main generator loop
        while True:
            # Check if we are in an error state (Prev or New)
            self.is_error = (
                True
                if (
                    (self.is_error is False)
                    and self.r.uniform(0, 1) <= self.error_rate_in_percentage
                )
                else self.is_error
            )

            # Manage error mode if needed
            if self.is_error:
                # If this is the first error step
                if self.steps == 0:
                    # Initialize error
                    self.current_error_length = int(
                        self.r.gauss(
                            mu=self.error_length_in_ticks,
                            sigma=0.1 * self.error_length_in_ticks,
                        )
                    )
                    self.scenario = {
                        metric_name: int(self.current_error_length * 0.1 * counter)
                        for counter, metric_name in enumerate(self.metrics.keys())
                    }

                    # Notify a metric to start an error state
                    self.notify_metric_of_error()

                    # Advance steps
                    self.steps += 1

                # If we are already in an error state, do we need to stop?
                elif self.steps == self.error_length_in_ticks:
                    # Change internal state
                    self.is_error = False
                    self.steps = 0
                    # Notify metrics
                    self.notify_metrics_of_normalization()

                # Normal in-error step
                else:
                    self.steps += 1

            # Generate a metric
            self.total_steps += 1
            metrics = {
                component_name: next(metric.generator())
                for component_name, metric in self.metrics.items()
            }
            new_metric = {}
            for metric_name, metric_values in metrics.items():
                for value_name, value in metric_values.items():
                    new_metric[
                        f"{metric_name}_{value_name}"
                        if value_name != "value"
                        else metric_name
                    ] = value
            yield new_metric



import faker
from ast import literal_eval as make_tuple
from random import Random


class LocationProvider(faker.providers.BaseProvider):
    """
    Creates locations for company deployments within given GPS Coordinates rectangle
    """

    def location(self, within: dict = {}):
        """

        :param within: GPS rectangle Coordinates containing:
                nw: ()
                se: ()
        :return: GPS Coordinate within the given rectangle
        """
        nw = make_tuple(str(within["nw"]))
        se = make_tuple(str(within["se"]))

        width = abs(nw[1] - se[1])
        height = abs(nw[0] - se[0])

        r = Random()

        longitude = se[0] + r.uniform(0, height)
        latitude = se[1] + r.uniform(0, width)

        location = (longitude, latitude)

        return location



from metrics_gen.deployment_generator import deployment_generator
from metrics_gen.metrics_generator import Generator_df
import pandas as pd
import yaml


def get_deployment(configuration: dict) -> pd.DataFrame:
    dep_gen = deployment_generator()
    deployment = dep_gen.generate_deployment(configuration=configuration)
    return deployment


class TestMetrics:
    configuration: dict = yaml.safe_load(
        open(
            "./tests/test_configuration.yaml",
            "r",
        )
    )
    metrics_configuration: dict = configuration.get("metrics", {})
    deployment: pd.DataFrame = get_deployment(configuration)

    def test_metric_as_dict(self):
        met_gen = Generator_df(self.configuration, user_hierarchy=self.deployment)
        generator = met_gen.generate(as_df=False)
        assert (generator, "No generated data was created")

    def test_metric_as_df(self):
        met_gen = Generator_df(self.configuration, user_hierarchy=self.deployment)
        generator = met_gen.generate(as_df=True)
        assert (generator, "No generated data was created")



from metrics_gen.deployment_generator import deployment_generator
import yaml


class TestDeployment:
    metrics_configuration: dict = yaml.safe_load(
        open(
            "./tests/test_configuration.yaml",
            "r",
        )
    )
    deployment_configuration: dict = metrics_configuration.get("deployment", {})

    @staticmethod
    def _get_deployment(configuration: dict = metrics_configuration):
        dep_gen = deployment_generator()
        deployment = dep_gen.generate_deployment(configuration=configuration)
        return deployment

    def test_deploy_with_add_levels(self):
        dep_gen = deployment_generator()
        faker = dep_gen.get_faker()

        for level, level_configs in self.deployment_configuration.items():
            devices_to_generate = level_configs.get("num_items", 1)
            level_type = level_configs.get("faker", "msisdn")
            dep_gen.add_level(
                level,
                number=devices_to_generate,
                level_type=getattr(faker, level_type),
            )
        deployment = dep_gen.generate_deployment()
        assert (deployment is not None, "Deployment should have values")

    def test_deploy_with_configuration_file(self):
        deployment = self._get_deployment()
        assert (deployment is not None, "Deployment should have values")

    def test_deploy_with_dep_configuration_only(self):
        deployment = self._get_deployment(self.deployment_configuration)
        assert (deployment is not None, "Deployment should have values")

    def test_deployment_get_faker(self):
        deployment = deployment_generator()
        faker = deployment.get_faker()

        assert (faker, "Faker not created")

    def test_all_columns_are_created(self):
        deployment = self._get_deployment()
        generated_levels = deployment.shape[1]

        num_levels = len(self.deployment_configuration.keys())
        assert (
            generated_levels == num_levels,
            f"Only {generated_levels} were created although {num_levels} were specified",
        )

    def test_number_of_generated_devices(self):
        deployment = self._get_deployment()
        generated_devices = deployment.shape[0]

        num_devices = 1
        for _, level_configs in self.deployment_configuration.items():
            num_devices = num_devices * level_configs.get("num_items", 1)
        assert (
            generated_devices == num_devices,
            f"Not enough devices created, {num_devices} should have been created while only {generated_devices} devices were created",
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="metrics_gen",
    version="0.1.1",
    author="Or Zilberman",
    author_email="zilbermanor@gmail.com",
    description="Fake deployment and data generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mlrun/metrics-gen",
    packages=setuptools.find_packages(exclude=("tests")),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
